package il.ac.tau.cs.software1.gamesearch;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

public class ExhaustiveGameSearch {

	private int depth;

	/**
	 * 
	 * @param depth
	 * @precondition depth > 0
	 */
	public ExhaustiveGameSearch(int depth) {
		this.depth = depth;
	}

	/**
	 * 
	 * @param b
	 *            the current state of the board
	 * @param white_move
	 *            true if it's a White move, false if a Black move
	 * @precondition if white_move==true, then the board's state should be such
	 *               that it is the White's turn, otherwise it should be the
	 *               black's turn.	                
	 * @precondition b.isGameOver() == false
	 * @return the optimal move for the player whose turn it is, given the
	 *         partial game tree with leaves that are either true leaves
	 *         (game-end situations) or this.depth moves down the tree. The
	 *         optimality is with respect to the min-max rules of the game tree
	 *         and to the score values at the leaves.
	 *         
	 */

    public ConnectFourMove optimalMove(ConnectFourBoard b, boolean white_move) {

        ConnectFourMove[] moves = b.legalMoves(); // must be nonempty
        ConnectFourMove best_move = null;

        double score;
        if (white_move) { // a minimizing tree node
          score = 1.0 / 0.0; // infinity
          for (int i = 0; i < moves.length; i++) {
            double iscore = exhaustiveSearch(b.move(moves[i], true), !white_move, 0);
            if (iscore < score) {
              score = iscore;
              best_move = moves[i];
            }
          }
        } else { // a maximizing tree node
          score = -1.0 / 0.0; // -infinity
          for (int i = 0; i < moves.length; i++) {
            double iscore = exhaustiveSearch(b.move(moves[i], false), !white_move, 0);
            if (iscore > score) {
              score = iscore;
              best_move = moves[i];
            }
          }
        }
      
        return best_move;
    }

    /**
     * A recursive helper method for optimalMove
     * 
     * @param b
     * @param white_move
     * @param d
     * @return
     */
    
    private double exhaustiveSearch(ConnectFourBoard b, boolean white_move, int d) {
      double s = b.isGameOver();
      if (s != -1.0) return s;
      
        if (d >= this.depth)
            return b.score();

        ConnectFourMove[] moves = b.legalMoves();
        double score;

        if (white_move) { // a minimizing tree node
          score = 1.0 / 0.0; // infinity
          for (int i = 0; i < moves.length; i++) {
                double iscore = exhaustiveSearch(b.move(moves[i], true), !white_move, d + 1);
                score = iscore < score ? iscore : score;
            }
        } else { // a minimizing tree node
          score = -1.0 / 0.0; // -infinity
          for (int i = 0; i < moves.length; i++) {
                double iscore = exhaustiveSearch(b.move(moves[i], false), !white_move, d + 1);
                score = iscore > score ? iscore : score;
            }
         }
        return score;
    }
    
    /**
     * 
     * @param moves  an array of moves
     * @param col    an integer (perhaps a move, perhaps not)
     * @precondition none
     * @return       true iff col is one of the moves in the moves array
     */

    private static boolean contains(ConnectFourMove[] moves, int col){
      for (int i=0; i<moves.length; i++) {
        if (moves[i].column()== col) return true;
      }
      return false;
  }

  /**
   * playInteractive
   * 
   * Creates a new game board and play either black or white; the user plays
   * the other side.
   * 
   */

  public void playInteractive(boolean computer_is_white) {
    String whiteName = computer_is_white ? "Computer" : "You";
    String blackName = computer_is_white ? "You" : "Computer";

    System.out.println("White player = " + whiteName);
    System.out.println("Black player = " + blackName);

    ConnectFourBoard b = new ConnectFourBoard();

    b.print();
    
    double s;
    boolean white_move = true;

    do {
      ConnectFourMove m;
      BufferedReader reader = new BufferedReader(new InputStreamReader(System.in));
          
      if (computer_is_white == white_move){
        // computer's turn;
        m = optimalMove(b, computer_is_white);
        b = b.move(m, computer_is_white);
      } else {
        ConnectFourMove[] legalMoves = b.legalMoves();
        System.out.print("Enter your move:");
              
        int col=-1;
        while(true){                    
          try {
            String inputStr = reader.readLine();
            try {
              col = Integer.decode(inputStr)-1;                           
            } catch(Exception ex) {}
            System.out.println();
          } catch (IOException e) {
            e.printStackTrace();
            System.exit(-1);
          }   
          if (contains(legalMoves, col)) break;
          System.out.print("Not a legal move. Enter a new move:");                    
        }
        m = new ConnectFourMove(col);
        b = b.move(m, !computer_is_white);
      }

      System.out.println(white_move ? "White Moves:" : "Black Moves:");
      b.print();
      white_move = !white_move;
      s = b.isGameOver();

    } while (s == -1.0);

    if (s == 0.0)
      System.out.println("White Wins!");
    if (s == 1.0)
      System.out.println("Black Wins!");
    if (s == 0.5)
      System.out.println("A Draw");

  }


  /**
   * @param args
   */
  public static void main(String[] args) {
      ExhaustiveGameSearch e = new ExhaustiveGameSearch(4);
      e.playInteractive(false);
  }
}
